@@ -3,14 +3,17 @@ from pathlib import Path
 import shutil
 import glob
 import click
+import subprocess
 import configparser
 import datetime
 import time
 import urllib.request
+
 import fileinput
 import cwltool.factory
 import cwltool.context
 
+
 @click.group()
 def cli():
    pass
@@ -43,16 +46,33 @@ def get_running_processes(ctx, args, incomplete):
     # Show running processes to be stopped
     return [os.system("ps ax | grep -v time | grep '[c]wl-runner' | awk '{print $1}'")]
 
+# Return the current dir where the command is run
+def getCurrentDir():
+    if os.name == 'nt':
+        # If Windows
+        return '$\{PWD\}'
+    else:
+        # Linux and MacOS
+        return '$(pwd)'
+
+# Change permissions to 777 recursively. 
+# See https://stackoverflow.com/questions/16249440/changing-file-permission-in-python
 def chmod777(path):
     for dirpath, dirnames, filenames in os.walk(path):
-        os.chmod(dirpath, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
-        # os.chmod(dirpath, 0o777)
-        # os.chmod('workspace/import', stat.S_IRWXO)
-        # shutil.chown(dirpath, owner)
+        # Handle permissions errors
+        try:
+            os.chmod(dirpath, stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
+            # os.chmod(dirpath, 0o777)
+            shutil.chown(dirpath, user=os.getuid(), group=os.getgid())
+        except:
+            click.echo(click.style('[d2s]', bold=True) + ' Issue while updating permissions for ' + dirpath + '.')
+        
         for filename in filenames:
-            os.chmod(os.path.join(dirpath, filename), stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
-            # os.chmod(os.path.join(dirpath, filename), 0o777)
-            # shutil.chown(os.path.join(dirpath, filename), owner)
+            try:
+                os.chmod(os.path.join(dirpath, filename), stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO)
+                shutil.chown(os.path.join(dirpath, filename), user=os.getuid(), group=os.getgid())
+            except:
+                click.echo(click.style('[d2s]', bold=True) + ' Issue while updating permissions for ' + os.path.join(dirpath, filename))
 
 @cli.command()
 @click.argument('projectname', nargs=1)
@@ -67,8 +87,6 @@ def init(ctx, projectname):
 
     config = configparser.ConfigParser()
     config['d2s'] = {}
-    # os.system('echo "UID=$UID" > .env')
-    # os.system('echo "GID=$GID" >> .env')
 
     click.echo(click.style('[d2s] ', bold=True) + 'You can generate a new project on GitHub using the provided template:')
     click.secho('https://github.com/MaastrichtU-IDS/d2s-transform-template/generate', bold=True)
@@ -83,19 +101,19 @@ def init(ctx, projectname):
     if not os.path.exists('./d2s-cwl-workflows'):
         os.system('git submodule add --recursive https://github.com/MaastrichtU-IDS/d2s-cwl-workflows.git')
 
-    # Create workspace directories
-    os.makedirs('workspace/output/tmp-outdir', exist_ok=True)
-    os.makedirs('workspace/output/tmp-outdir', exist_ok=True)
-    os.makedirs('workspace/import', exist_ok=True)
-    os.makedirs('workspace/logs', exist_ok=True)
-    os.makedirs('workspace/dumps/rdf/releases/1', exist_ok=True)
-    os.makedirs('workspace/dumps/hdt', exist_ok=True)
-    # Linux only, use os.chmod
-    # os.chmod(path, stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH)
-    # https://stackoverflow.com/questions/16249440/changing-file-permission-in-python
-    chmod777('workspace/import')
-    # Read, write, and execute by others
-    # os.system('chmod -R 777 workspace/import')
+    # Create workspace directories and chmod 777
+    listToCreate = ["input", "output", "import", "output/tmp-outdir", 
+        "logs", "dumps/rdf/releases/1", 'dumps/hdt', 'virtuoso', 'tmp-virtuoso']
+    click.echo(click.style('[d2s]', bold=True) + ' Creating following folders in workspace: ' + ", ".join(listToUpdate))
+    for fileToCreate in listToCreate:
+        os.makedirs('workspace/' + fileToCreate, exist_ok=True)
+    chmod777('workspace')
+    # Copy load.sh in workspace for Virtuoso bulk load
+    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/virtuoso/load.sh')
+    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/tmp-virtuoso/load.sh')
+    # TODO: improve this to include it in Docker deployment
+
+
     # Get RMLStreamer from home dir to qvoid download each time
     user_home_dir = str(Path.home())
     if os.path.exists(user_home_dir + '/RMLStreamer.jar'):
@@ -103,16 +121,8 @@ def init(ctx, projectname):
     else:
         click.echo(click.style('[d2s]', bold=True) + ' Downloading RMLStreamer.jar... [80M]')
         urllib.request.urlretrieve ("https://github.com/vemonet/RMLStreamer/raw/fix-mainclass/target/RMLStreamer-1.2.2.jar", "workspace/RMLStreamer.jar")
-    # os.system('wget -a workspace/RMLStreamer.jar https://github.com/vemonet/RMLStreamer/raw/fix-mainclass/target/RMLStreamer-1.2.2.jar')
     chmod777('workspace/RMLStreamer.jar')
 
-    # Copy load.sh in workspace for Virtuoso bulk load
-    os.makedirs('workspace/virtuoso', exist_ok=True)
-    os.makedirs('workspace/tmp-virtuoso', exist_ok=True)
-    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/virtuoso/load.sh')
-    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/tmp-virtuoso/load.sh')
-    # TODO: improve this to include it in Docker deployment
-
     click.echo()
     # Copy GraphDB zip file to the right folder in d2s-cwl-workflows
     click.echo(click.style('[d2s]', bold=True) + ' The GraphDB triplestore needs to be downloaded for licensing reason.\n'
@@ -121,13 +131,12 @@ def init(ctx, projectname):
         + ' and provide your email to receive the URL to download ' + click.style('GraphDB version 9.1.1 standalone zip', bold=True))
     
     graphdb_path = click.prompt(click.style('[?]', bold=True) + ' Enter the path to the GraphDB distribution 9.1.1 zip file used to build the Docker image. Default', default='~/graphdb-free-9.1.1-dist.zip')
-    # os.system('cp ' + graphdb_path + ' ./d2s-cwl-workflows/support/graphdb')
+    # Get GraphDB installation file
     if os.path.exists(graphdb_path):
         shutil.copyfile(graphdb_path, './d2s-cwl-workflows/support/graphdb')
     else:
         click.echo(click.style('[d2s]', bold=True) + ' GraphDB installation file not found. Copy the zip file in d2s-cwl-workflows/support/graphdb after download.')
 
-    
     with open('.d2sconfig', 'w') as configfile:
         config.write(configfile)
     
@@ -151,13 +160,21 @@ def init(ctx, projectname):
 def update(services, permissions):
     """Update Docker images"""
     if permissions:
-        click.echo(click.style('[d2s]', bold=True) + ' Password will be asked to updates following folder permissions in workspace: input, output, import, tmp-virtuoso')
-        os.system('sudo chmod -R 777 workspace/input')
-        os.system('sudo chmod -R 777 workspace/output')
-        os.system('sudo chmod -R 777 workspace/import')
-        os.system('sudo chmod -R 777 workspace/dumps')
-        os.system('sudo chmod -R 777 workspace/tmp-virtuoso')
+        listToUpdate = ["input", "output", "import", "dumps", "tmp-virtuoso", "virtuoso"]
+        click.echo(click.style('[d2s]', bold=True) + ' Password might be asked to updates following folder permissions in workspace: ' + ", ".join(listToUpdate))
+        for fileToUpdate in listToUpdate:
+            chmod777('workspace/' + fileToUpdate)
+            os.makedirs('workspace/' + fileToUpdate, exist_ok=True)
+        # Copy load.sh in workspace for Virtuoso bulk load
+        shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/virtuoso/load.sh')
+        shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/tmp-virtuoso/load.sh')
+        chmod777('workspace/virtuoso/load.sh')
+        chmod777('workspace/tmp-virtuoso/load.sh')
+        # TODO: improve this to include it in Docker deployment
+        click.echo(click.style('[d2s]', bold=True) + ' Most permissions issues can be fixed by changing the owner of the file while running as administrator')
+        click.secho('sudo chown -R ' + os.getuid() + ':' + os.getgid() + ' workspace', bold=True)
     else:
+        # Update Docker images (pull and build graphdb)
         services_string = " ".join(services)
         os.system(docker_compose_cmd + 'pull ' + services_string)
         if not services or "graphdb" in services:
@@ -207,13 +224,21 @@ def start(services, deploy):
     # Ask user to create the GraphDB test repository
     click.echo(click.style('[d2s] ', bold=True) + services_string + ' started.')
     if 'graphdb' in services or 'demo' in services:
-        if click.confirm(click.style('[?]', bold=True) + ' Do you want to create the ' 
-        + click.style('demo repository', bold=True) + ' in GraphDB?'):
-            click.echo(click.style('[d2s] ', bold=True) 
-                + 'Creating the repository, it should take about 20s.')
-            time.sleep(10)
-            # TODO: use urllib
-            os.system('curl -X POST http://localhost:7200/rest/repositories -F "config=@d2s-cwl-workflows/support/graphdb-repo-config.ttl" -H "Content-Type: multipart/form-data"')
+        click.echo(click.style('[d2s] ', bold=True) 
+                + 'Create repository on GraphDB: http://localhost:7200/repository')
+        # TODO: Creating GraphDB repo by posting multiform with urllib needs to be changed
+        # if click.confirm(click.style('[?]', bold=True) + ' Do you want to create the ' 
+        # + click.style('demo repository', bold=True) + ' in GraphDB?'):
+        #     click.echo(click.style('[d2s] ', bold=True) 
+        #         + 'Creating the repository, it should take about 20s.')
+        #     time.sleep(10)
+        #     localGraphdbUrl = 'http://localhost:7200/rest/repositories'
+        #     headers = {'Content-Type': 'multipart/form-data'}
+        #     request = urllib.request.Request(localGraphdbUrl, 
+        #         open('d2s-cwl-workflows/support/graphdb-repo-config.ttl', 'rb'),
+        #                             headers=headers)
+        #     response = urllib.request.urlopen(request)
+        #     os.system('curl -X POST http://localhost:7200/rest/repositories -F "config=@d2s-cwl-workflows/support/graphdb-repo-config.ttl" -H "Content-Type: multipart/form-data"')
     click.echo()
     click.echo(click.style('[d2s]', bold=True) 
         + ' You can now download data to run a first workflow:')
@@ -305,9 +330,9 @@ def download(datasets):
 def rml(dataset, detached, mapper, parallelism):
     """Run RML Streamer"""
     if (detached):
-        detached_arg = "-d"
+        detached_arg = '-d'
     else:
-        detached_arg = "-it"
+        detached_arg = '-it'
 
     for file in os.listdir('./datasets/' + dataset + '/mapping'):
      mapping_filename = os.fsdecode(file)
@@ -316,24 +341,30 @@ def rml(dataset, detached, mapper, parallelism):
         mapping_filepath = 'datasets/' + dataset + '/mapping/' + mapping_filename
         click.echo(click.style('[d2s]', bold=True) + ' Execute mappings from ' 
             + click.style(mapping_filepath, bold=True))
-    
-        # TODO: store logs in a file and get the run time
+
+        output_filename = 'rml'
         if mapper:
             output_filename = 'rmlmapper-' + mapping_filename.replace('.', '_') + '-' + dataset + '.nt'
-            rmlstreamer_cmd = 'docker run ' + detached_arg + ' -v $(pwd)/workspace:/mnt/workspace -v $(pwd)/datasets:/mnt/datasets umids/rmlmapper:4.7.0 -m /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' -o /mnt/workspace/import/' + output_filename
+            rml_cmd = 'docker run ' + detached_arg + ' -v ' + getCurrentDir() + '/workspace:/mnt/workspace -v ' + getCurrentDir() + '/datasets:/mnt/datasets umids/rmlmapper:4.7.0 -m /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' -o /mnt/workspace/import/' + output_filename
         else:
             # Use RMLStreamer
             output_filename = 'rmlstreamer-' + mapping_filename.replace('.', '_') + '-' + dataset + '.nt'
-            rmlstreamer_cmd = 'docker exec ' + detached_arg + ' d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -c io.rml.framework.Main /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' --outputPath /mnt/workspace/import/' + output_filename + ' --job-name "[d2s] RMLStreamer ' + mapping_filename + ' - ' + dataset + '"'
-            print(rmlstreamer_cmd)
+            rml_cmd = 'docker exec ' + detached_arg + ' d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -c io.rml.framework.Main /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' --outputPath /mnt/workspace/import/' + output_filename + ' --job-name "[d2s] RMLStreamer ' + mapping_filename + ' - ' + dataset + '"'
+            click.echo(click.style('[d2s]', bold=True) + ' Check the jobs running at ' 
+                + click.style('http://localhost:8078/#/job/running', bold=True))
             ## Try parallelism:
-            # rmlstreamer_cmd = 'docker exec ' + detached_arg + ' d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -p ' + parallelism + ' -c io.rml.framework.Main /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' --outputPath /mnt/workspace/import/' + output_filename + ' --job-name "[d2s] RMLStreamer ' + mapping_filename + ' - ' + dataset + '" --parallelism ' + parallelism + ' --enable-local-parallel'
+            # rml_cmd = 'docker exec ' + detached_arg + ' d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -p ' + parallelism + ' -c io.rml.framework.Main /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/' + mapping_filename + ' --outputPath /mnt/workspace/import/' + output_filename + ' --job-name "[d2s] RMLStreamer ' + mapping_filename + ' - ' + dataset + '" --parallelism ' + parallelism + ' --enable-local-parallel'
     
-        os.system(rmlstreamer_cmd)
+        os.system(rml_cmd)
+        # TODO: store logs in a file and get the run time. Use subprocess.call
+        # with open('workspace/logs/' + output_filename + '.log', 'a') as output:
+        #     process = subprocess.call(rml_cmd, shell=True, 
+        #         stdout=output, stderr=output)
+
+        # click.echo(click.style('[d2s]', bold=True) + ' PID of running process: ' + str(process.pid))
         click.echo(click.style('[d2s]', bold=True) + ' Output file in ')
         click.secho('workspace/import/' + output_filename, bold=True)
-        click.echo(click.style('[d2s]', bold=True) + ' Check the jobs running at ' 
-                + click.style('http://localhost:8078/#/job/running', bold=True))
+    
     # Try parallelism:
     # rmlstreamer_cmd = 'docker exec -d d2s-cwl-workflows_rmlstreamer_1 /opt/flink/bin/flink run -p 4 /mnt/workspace/RMLStreamer.jar --path /mnt/datasets/' + dataset + '/mapping/rml-mappings.ttl --outputPath /mnt/workspace/import/rml-output-' + dataset + '.nt --job-name "[d2s] RMLStreamer ' + dataset + '" --enable-local-parallel'
     # Try to use docker-compose, but exec dont resolve from the -f file
@@ -359,11 +390,11 @@ def run(workflow, dataset, get_mappings, detached):
 
     # TODO: Trying to fix issue where virtuoso bulk load only the first dataset we run
     # It needs restart to work a second time
-    click.echo(click.style('[d2s] ', bold=True) 
-        + 'Restart tmp Virtuoso and delete file in '
-        + click.style('workspace/output', bold=True))
-    os.system(docker_compose_cmd + 'stop tmp-virtuoso')
-    os.system(docker_compose_cmd + 'up -d --force-recreate tmp-virtuoso')
+    # click.echo(click.style('[d2s] ', bold=True) 
+    #     + 'Restart tmp Virtuoso and delete file in '
+    #     + click.style('workspace/output', bold=True))
+    # os.system(docker_compose_cmd + 'stop tmp-virtuoso')
+    # os.system(docker_compose_cmd + 'up -d --force-recreate tmp-virtuoso')
     # TODO: fix this dirty Virtuoso deployment 
     # Virtuoso unable to handle successive bulk load + permission issues + load.sh in the virtuoso containers
     # I don't know how, they managed to not put it in the container... They had one job...
@@ -372,13 +403,9 @@ def run(workflow, dataset, get_mappings, detached):
     shutil.rmtree('workspace/output', ignore_errors=True, onerror=None)
     for file in glob.glob("workspace/tmp-virtuoso/*.nq"):
         os.remove(file)
-    # Make sure the load.sh script is in the tmp Virtuoso folder
-    os.system('mkdir -p workspace/tmp-virtuoso && cp d2s-cwl-workflows/support/virtuoso/load.sh workspace/tmp-virtuoso')
-    os.makedirs('workspace/tmp-virtuoso', exist_ok=True)
-    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/virtuoso/load.sh')
-    shutil.copyfile('d2s-cwl-workflows/support/virtuoso/load.sh', 'workspace/tmp-virtuoso/load.sh')
-    
+
     if (detached):
+        # TODO: Find a better solution to work on windows
         cwl_command = 'nohup time '
     else:
         cwl_command = ''
@@ -404,18 +431,7 @@ def run(workflow, dataset, get_mappings, detached):
         click.echo()
         click.echo(click.style('[d2s] ', bold=True) + 'Browse the file generated by the workflow in ' 
             + click.style('workspace/output/' + dataset, bold=True))
-        # click.echo(click.style('[d2s] ', bold=True)
-        #     + 'Access the linked data browser for Virtuoso at '
-        #     + click.style('http://localhost:8891', bold=True))
-        # click.echo(click.style('[d2s] ', bold=True)
-        #     + 'Access Virtuoso (temp store) at '
-        #     + click.style('http://localhost:8890', bold=True))
-        # click.echo(click.style('[d2s] ', bold=True) 
-        #     + 'Access the linked data browser for GraphDB at '
-        #     + click.style('http://localhost:7201', bold=True))
-        # click.echo(click.style('[d2s] ', bold=True) 
-        #     + 'Access GraphDB at '
-        #     + click.style('http://localhost:7200', bold=True))
+
     if (detached):
         click.echo()
         click.echo(click.style('[d2s] ', bold=True) 
@@ -493,6 +509,7 @@ def dataset():
         + click.style(dataset_id + ' dataset', bold=True) 
         + ' has been generated')
     click.echo(click.style('[d2s]', bold=True) + ' Start edit them in ' + click.style('datasets/' + dataset_id, bold=True))
+    
     # Will not work on all platforms:
     # if click.confirm(click.style('[?]', bold=True) + ' Do you want to open the ' 
     #     + click.style('download', bold=True) + ' file to edit it?'):
